"""Common definitions for all test sequencies"""
import os
import socket
from test_runner.test_position import TestPosition
from test_runner.dut import Dut
from test_runner.test_case import FlowControl

TEST_POSITIONS = [
    TestPosition('1', 'left'),
    TestPosition('2', 'middle'),
    TestPosition('3', 'right'),
]

DB_HANDLER_NAME = 'test_data_db'

# FlowControl.CONTINUE keeps testing even some test fails
# FlowControl.STOP_ON_FAIL stops on first fail

FLOW_CONTROL = FlowControl.CONTINUE

# Wait DUT serial number from UI
SN_FROM_UI = True


# List of test sequences' directory names
TEST_SEQUENCES = os.listdir('test_definitions/sequences')


def parse_dut_info(info, test_position):
    return Dut(serial_number=info['sn'], test_position=test_position)


def get_my_ip(ext_url):
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as sock:
        sock.connect(ext_url)
        return sock.getsockname()[0]


# Use here the external url that want's to connect to to Iris
IRIS_IP = get_my_ip(("8.8.8.8", 80))



import datetime
import pathlib
import logging
from pymongo import MongoClient


class DatabaseHandler:
    def __init__(self, db_connection_string, db_name):
        self.db_client = MongoClient(db_connection_string, serverSelectionTimeoutMS=10000)
        self.db_name = db_name
        self.logger = logging.getLogger("DatabaseHandler")

    def get_media_file_path(self, file_name):
        return self.db_client[self.db_name].file_attachments.find_one({'name': file_name})[
            'file_path'
        ]

    def get_statistics(self):
        return {
            'yield': self.get_yield(),
            'uph': self.get_uph(),
        }

    def get_yield(self):

        pass_count = (
            self.db_client[self.db_name]
            .test_reports.find(
                {
                    'start_time': {'$gt': datetime.datetime.now() - datetime.timedelta(weeks=4)},
                    'overallResult': True,
                }
            )
            .count()
        )
        all_count = (
            self.db_client[self.db_name]
            .test_reports.find(
                {'start_time': {'$gt': datetime.datetime.now() - datetime.timedelta(weeks=4)}}
            )
            .count()
        )

        if all_count == 0:
            return 0

        return pass_count / all_count * 100

    def get_uph(self):
        return (
            self.db_client[self.db_name]
            .test_reports.find(
                {'start_time': {'$gt': datetime.datetime.now() - datetime.timedelta(hours=1)}}
            )
            .count()
        )

    def clean_db(self):
        # Get files older than one week
        file_delete_filter = {
            'added': {'$lt': datetime.datetime.now() - datetime.timedelta(weeks=1)}
        }
        old_files = self.db_client[self.db_name].file_attachments.find(file_delete_filter)

        # Delete files on HDD
        for _file in old_files:
            try:
                pathlib.Path(_file['file_path']).unlink()
            except FileNotFoundError:
                self.logger.warning(
                    "Trying to delete non-existent file attachment: %s", _file['file_path']
                )

        # Delete file entries on database
        self.db_client[self.db_name].file_attachments.delete_many(file_delete_filter)

        # Delete old test reports
        self.db_client[self.db_name].test_reports.delete_many(
            {'start_time': {'$lt': datetime.datetime.now() - datetime.timedelta(weeks=4)}}
        )

    def get_search_bar_items(self):
        return {
            'searchBarItems': [
                {
                    'name': 'limit_N',
                    'placeholder_txt': "Number of items to get",
                    'label': 'Number of items',
                    'type': 'textbox',
                },
                {
                    'name': 'limit_hours',
                    'placeholder_txt': "Hours",
                    'label': 'Hours',
                    'type': 'textbox',
                },
                {
                    'name': 'dut_identifier',
                    'placeholder_txt': "ABCD-1234",
                    'label': 'DUT ID',
                    'type': 'textbox',
                },
                {
                    'name': 'only_fails',
                    'placeholder_txt': "",
                    'label': 'Only Fails',
                    'type': 'checkbox',
                },
            ]
        }

    def search_db(self, search_args):

        db_filter = {}
        if 'limit_hours' in search_args:
            db_filter.update(
                {
                    'start_time': {
                        '$gt': datetime.datetime.now()
                        - datetime.timedelta(hours=search_args['limit_hours'])
                    }
                }
            )

        if 'dut_identifier' in search_args:
            db_filter.update({'serialnumber': search_args['dut_identifier']})

        if 'only_fails' in search_args:
            db_filter.update({'result': False})

        limit = 500

        if 'limit_N' in search_args:
            limit = search_args['limit_N']

        reps = (
            self.db_client[self.db_name]
            .test_reports.find(db_filter)
            .sort('start_time')
            .limit(limit)
        )

        return {'test_runs': list(reps)}

    def store_test_data_file_to_db(self, **kwargs):
        self.db_client[self.db_name].file_attachments.insert_one({**kwargs})



LIMITS = {}

LIMITS["Second"] = {
    "Measurement1": {
        "limit": lambda measurement: 90 < measurement < 100,
        "unit": "dB",
        "report_limit": "Overriden representation of limit to be written to the report.",
    },
    "Measurement2": {"limit": lambda measurement: measurement == [123, 456, 1, 3]},
    # Measurement result may be something else than boolean
    # Boolean True is only result that is reported as "Pass",
    # All other results are considered "Fail"
    "Measurement3": {"limit": lambda measurement: True if measurement < 100 else "ABC"},
    "Measurement4": {"limit": lambda measurement: measurement.contains("123abc")},
    "Measurement5": {"limit": lambda measurement: complicated_limit_definition(measurement)},
}

LIMITS["First"] = {
    "Measurement1": {"limit": lambda measurement: 90 < measurement < 100},
    "Measurement2": {"limit": lambda measurement: measurement == [123, 456, 1, 3]},
}


def complicated_limit_definition(measurement):
    # Must return true/false
    pass



from test_runner.test_case import TestCase


class First(TestCase):
    def pre_test(self):
        # pre_test() will be run always before test().
        # Use "First_pre" to specify when pre_test of this class is started.
        # If "First_pre" is not specified, pre_test will be run just before the test.
        pass

    def test(self):
        self.logger.info("At first test case")
        self.show_operator_instructions("Ínstructions for the operator")
        with open('test_data.txt', 'w') as test_data_file:
            test_data_file.write("Sample data inside the file")

        self.store_test_data_file(
            'test_data.txt',
            'Measurement1_data',
            any_additional_data="Is added as parameter",
            one_more_parameter="sample parameter",
        )

        self.new_measurement("Measurement1", 123)
        # instruments["G5"].do_something_with_instrument()
        self.new_measurement("Measurement2", [123, 456, 1, 3])

        import time

        time.sleep(2)

    def post_test(self):
        # post_test will be run always on background after the test
        pass



import datetime
import sys
import os
import json
import importlib
import threading
from test_runner import exceptions


def import_by_name(name, error_message, logger):
    sys.path.append(os.getcwd())
    sys.path.append(os.path.join(os.getcwd(), 'test_definitions'))

    try:
        imp = importlib.import_module(name)

        if 'common' not in name:
            importlib.reload(imp)

        # TODO: This hides also errors on nested imports
    except ImportError as err:
        logger.warning(err)
        logger.warning(error_message)
        sys.exit(-1)

    return imp


def get_test_pool_definitions(logger):
    """Returns test definition pool"""

    return import_by_name('test_case_pool', "test_case_pool missing?", logger)


def get_test_definitions(sequence_name, logger):
    """Returns test definitions"""

    err = (
        "Error loading "
        + sequence_name
        + ". Remember, you can create new definition template with --create argument."
    )

    return import_by_name("sequences." + sequence_name, err, logger)



from test_runner.test_case import TestCase


class Second(TestCase):
    def pre_test(self):
        pass

    def test(self):
        self.logger.info("At second test case")
        self.new_measurement("Measurement1", 123)
        # instruments["G5"].do_something_with_instrument()
        self.new_measurement("Measurement2", [123, 456, 1, 3])
        import time
        time.sleep(2)

    def post_test(self):
        pass



"""Defines tests and their order"""

# Set order of tests
# Use "_pre" to specify when pre-test is started
# If "_pre" is not specified, pre-test will be run before the test
# Post-test will be run always on background after the test
TESTS = ["Second_pre", "First", "Second", "Third", "Fourth", "PoolTestCase"]

SKIP = ["Third", "Fourth"]



import logging
import asyncio
import tornado
import tornado.websocket


class WebsocketStreamHandler(logging.StreamHandler):
    def __init__(self, websocket):
        super().__init__(self)
        self.websocket = websocket
        self.loop = websocket.loop

    def emit(self, record):
        self.websocket.send(record)

    def write(self, record):
        record = record.rstrip()
        if record:  # and: not self.websocket.closed:
            self.loop.call_soon_threadsafe(self.websocket.write_message, record)

    def flush(self):
        pass


class LogsWebSocketHandler(tornado.websocket.WebSocketHandler):
    """
    Note that Tornado uses asyncio. Since we are using threads on our backend
    we need to use call_soon_threadsafe to get messages through.
    """

    def __init__(self, application, request, **kwargs):

        self._root_logger = logging.getLogger()
        self._stream_handler = None
        super().__init__(application, request, **kwargs)
        self.loop = asyncio.get_event_loop()  # pylint: disable=W0201

    def open(self, *args, **kwargs):
        """Called when websocket is opened"""

        # Get websocket logger just for getting logging settings
        ws_logger = logging.getLogger('websocket')
        # Get handler which is configured is logging.yaml
        handler = list(
            filter(
                lambda x: x.name is not None and x.name.lower() == 'websocket', ws_logger.handlers
            )
        )[0]

        weblogger = WebsocketStreamHandler(self)
        self._stream_handler = logging.StreamHandler(weblogger)
        self._stream_handler.formatter = handler.formatter
        self._stream_handler.level = handler.level

        self._root_logger.addHandler(self._stream_handler)

        self._root_logger.info('Websocket logger connected')

    def on_close(self):
        """Called when websocket is closed"""
        self._root_logger.removeHandler(self._stream_handler)

    def on_message(self, message):
        """Called when message comes from client through websocket"""
        # self.write_message("echo: %r" % message)

    def check_origin(self, origin):  # pylint: disable=R0201, W0613
        """Checks whether websocket connection from origin is allowed.

        We will allow all connection which is actually potential safety risk. See:
        https://www.tornadoweb.org/en/stable/websocket.html#tornado.websocket.WebSocketHandler.check_origin
        """
        return True

    def data_received(self, chunk):
        pass



"""Example of test case defined at 'pool'. These cases are available
for all test sequences"""

from test_runner.test_case import TestCase


class PoolTestCase(TestCase):
    def pre_test(self):
        pass

    def test(self):
        self.logger.info("Running test case from pool")
        self.logger.info(
            "Parameters can be used to parametrize the testing: "
            + self.parameters["param1"]
        )

    def post_test(self):
        pass



class Dut:
    def __init__(self, serial_number, test_position, hw_id=None, patch=None, additional_info=None):
        self.test_position = test_position
        self.serial_number = serial_number
        self.hw_id = hw_id
        self.patch = patch
        self.additional_info = additional_info
        self.test_cases = {}
        self.pass_fail_result = 'testing'
        self.failed_steps = []
        self.error_steps = []

    def get_dut_dict(self):

        return {
            'test_position': self.test_position,
            'serial_number': self.serial_number,
            'hw_id': self.hw_id,
            'patch': self.patch,
            'additional_info': self.additional_info,
            'test_cases': self.test_cases,
            'result': self.pass_fail_result,
            'failed_steps': self.failed_steps,
            'error_steps': self.error_steps,
        }



import base64
import os
import bcrypt

# This is our "user database". As you see passwords and users are hardcoded.
USER_ROLES = {
    "Service": '$2b$12$9krtCz3xld.d4JRI2TYixuxhaNiRXysFtiGSPdYzGQ/ilxzJIbHHK',
    "Admin": '$2b$12$6ov3vhUecjQJGHqWsAI/vOHK.PBvrLTY0ZdhZUcvCrZtJyuebM2se',
}


def authenticate(user, password):
    if user in ["Service", "Admin"]:
        if bcrypt.checkpw(password.encode(), USER_ROLES[user].encode()):
            return True

    if user == 'Operator':
        return True
    return False


def get_cookie_secret():
    '''Get a number that should be secure / random enough for our cryptographic needs.'''

    try:
        with open("cookie_secret") as f:
            secret = f.read()
    except FileNotFoundError:
        with open("cookie_secret", 'w') as f:
            secret = base64.b64encode(os.urandom(50)).decode('ascii')
            f.write(secret)
    return secret


if __name__ == "__main__":
    pass



class IrisError(Exception):
    """Base class for exceptions in this module."""

    pass


class InputError(IrisError):
    # TODO: modify this to define on which step error happened etc.
    # And give proper name
    """Exception raised for errors in the input.

    Attributes:
        expression -- input expression in which the error occurred
        message -- explanation of the error
    """

    def __init__(self, expression, message):
        self.expression = expression
        self.message = message


class TestCaseNotFound(IrisError):
    """Raised when test case is not found"""

    pass



"""Base for all test cases"""
from abc import ABC, abstractmethod
import inspect
import logging
import datetime
import time

from enum import Enum
from pymongo import MongoClient
from pathlib import Path


class FlowControl(Enum):
    """Defines flow control options"""

    """Stop if result was fail"""
    STOP_ON_FAIL = 1

    """Continue always"""
    CONTINUE = 3


class TestCase(ABC):
    """Base for all test cases"""

    def __init__(self, limits, report_progress, dut, parameters, db_handler, common_definitions):
        self.flow_control = common_definitions.FLOW_CONTROL
        self.logger = logging.getLogger('test_case')
        self.parameters = parameters
        self.instruments = common_definitions.INSTRUMENTS
        self.dut = dut
        self.report_progress = report_progress
        self.limits = limits
        self.logger = logging.getLogger('test_case')
        self.start_time = None
        self.start_time_monotonic = None
        self.duration_s = None
        self.end_time = None
        self.test_position = None
        self.name = self.__class__.__name__
        self.db_handler = db_handler
        self.my_ip = common_definitions.IRIS_IP
        # Initialize measurement dictionary
        self.dut.test_cases[self.name] = {'id': id(self), 'result': 'testing', 'measurements': {}}

    def show_operator_instructions(self, message, append=False):
        self.report_progress.show_operator_instructions(message, append)

    @abstractmethod
    def test(self):
        """Defines the test case."""

    def pre_test(self):
        """This method will be run on background before the actual test"""

    @abstractmethod
    def post_test(self):
        """This method will be run on background after the actual test"""

    def clean(self):
        """Called after all steps are done"""

    def clean_error(self):
        """Called after all steps are done"""

    def clean_pass(self):
        """Called after all steps are done if pass_fail_result is True"""

    def clean_fail(self):
        """Called after all steps are done if pass_fail_result is not True"""

    def new_measurement(self, name, measurement):
        """Adds new measurement to measurement array"""
        self.logger.debug("New measurement: %s:%s:%s", self.name, name, measurement)

        if name not in self.dut.test_cases[self.name]['measurements']:
            self.dut.test_cases[self.name]['measurements'][name] = {}

        self.dut.test_cases[self.name]['measurements'][name]['measurement'] = measurement

    def stop_testing(self):
        """Stops testing before going to next test step"""
        self.test_position.stop_testing = True

    def result_handler(self, error=None):
        """Checks if test is pass or fail. Can be overridden if needed."""

        case = self.dut.test_cases[self.name]

        for measurement_name, measurement_dict in self.dut.test_cases[self.name][
            'measurements'
        ].items():
            pass_fail_result = 'pass'
            limit = None
            unit = None
            try:
                case['measurements'][measurement_name]["error"] = None

                measurement_value = measurement_dict['measurement']

                if self.name in self.limits:
                    limit = self.limits[self.name][measurement_name].get(
                        'report_limit',
                        inspect.getsource(self.limits[self.name][measurement_name]['limit']),
                    )

                    pass_fail_result = 'pass' if self.limits[self.name][measurement_name]['limit'](
                        measurement_value
                    ) else 'fail'

                    unit = self.limits[self.name][measurement_name].get('unit', '')

                case['measurements'][measurement_name]["unit"] = unit
                case['measurements'][measurement_name]["limit"] = limit
                case['measurements'][measurement_name]["result"] = pass_fail_result

            except Exception as exp:
                case['result'] = 'error'
                case['measurements'][measurement_name]["unit"] = unit
                case['measurements'][measurement_name]["limit"] = limit
                case['measurements'][measurement_name]["result"] = "ErrorOnLimits"
                case['measurements'][measurement_name]["error"] = str(type(exp)) + ': ' + str(exp)

            finally:
                self.set_pass_fail_result(pass_fail_result)


        if error:

            case['result'] = 'error'
            case['error'] = error
            self.set_pass_fail_result('error')


    def set_pass_fail_result(self, result):
        # Result must be set for both test case and DUT
        # If any of the test cases is fail/error, the result of DUT is fail/error.

        def __set_result(target):
            if target == 'error':
                # If result is set to error, it must stay on error
                pass
            elif target == 'fail':
                if result == 'error':
                    # From 'fail' go only to 'error'
                    target = result
            else:
                # From all other results ('pass', 'NA') go to any other result
                target = result

            return target

        # Set result for test case
        new_result = __set_result(self.dut.test_cases[self.name]['result'])
        self.dut.test_cases[self.name]['result'] = new_result

        # Set result for DUT
        if not hasattr(self, 'pass_fail_result'):
            self.dut.pass_fail_result = result
        else:
            self.dut.pass_fail_result = __set_result(self.dut.pass_fail_result)

        if result != 'pass':
            if self.flow_control == FlowControl.STOP_ON_FAIL:
                self.stop_testing()

            if result == 'fail':
                if self.name not in self.dut.failed_steps:
                    self.dut.failed_steps.append(self.name)

            if result == 'error':
                if self.name not in self.dut.error_steps:
                    self.dut.error_steps.append(self.name)


    def check_measurements_vs_limits(self):
        if self.name in self.limits:

            for limit in self.limits[self.name]:

                if 'optional' in self.limits[self.name] and self.limits[self.name]['optional']:
                    continue

                if (
                    self.name in self.dut.test_cases
                    and limit not in self.dut.test_cases[self.name]['measurements'].keys()
                ):

                    self.dut.pass_fail_result = 'error'

                    self.dut.test_cases[self.name]['result'] = 'error'
                    self.dut.test_cases[self.name]['error'] = f'Measurement "{limit}" missing'

                    if self.flow_control == FlowControl.STOP_ON_FAIL:
                        self.stop_testing()

    def run_pre_test(self):
        self.logger.debug("Running pre_test at %s", self.name)
        self.start_time = datetime.datetime.now()
        self.start_time_monotonic = time.monotonic()
        self.pre_test()
        self.evaluate_results()
        self.logger.debug("Pre_test done at %s", self.name)

    def run_test(self):
        self.logger.debug("Running test at %s", self.name)
        self.test()
        self.evaluate_results()
        self.logger.debug("Test done at %s", self.name)

    def run_post_test(self):
        self.logger.debug("Running post_test done at %s", self.name)
        self.post_test()

        self.evaluate_results()
        self.check_measurements_vs_limits()
        if self.dut.pass_fail_result:
            self.clean_pass()
        else:
            self.clean_fail()

        self.clean()

        self.end_time = datetime.datetime.now()
        self.duration_s = time.monotonic() - self.start_time_monotonic

        self.dut.test_cases[self.name].update(
            {
                "start_time": self.start_time,
                "end_time": self.end_time,
                "duration_s": round(self.duration_s, 2),
            }
        )
        self.logger.debug("Post_test done at %s", self.name)

    def handle_error(self, error):

        self.logger.warn("Error at %s: %s", self.name, error)
        self.result_handler(error)

        self.end_time = datetime.datetime.now()
        self.duration_s = time.monotonic() - self.start_time_monotonic

        self.dut.test_cases[self.name].update(
            {
                "start_time": self.start_time,
                "end_time": self.end_time,
                "duration_s": round(self.duration_s, 2),
            }
        )
        self.clean_error()

    def evaluate_results(self):

        self.result_handler()

    def store_test_data_file(self, source_file_path, dest_name, **kwargs):
        from common.test_report_writer import create_report_path

        unique_dest_name = (
            self.name + '_' + self.dut.serial_number + '_' + self.test_run_id + '_' + dest_name
        )

        report_path = create_report_path()
        dest_path = Path(report_path, 'file_attachments')
        dest_path.mkdir(parents=True, exist_ok=True)
        dest_path = dest_path / unique_dest_name
        Path(source_file_path).rename(dest_path)

        data = {
            'name': dest_name,
            'file_path': str(dest_path),
            'testRunId': self.test_run_id,
            'testCase': self.name,
            'dut': self.dut.serial_number,
            'added': datetime.datetime.now(),
            'size': Path(dest_path).stat().st_size,
            'expires': datetime.datetime.now() + datetime.timedelta(weeks=1),
            'url': f'http://{self.my_ip}/api/media/{dest_name}',
            **kwargs,
        }

        self._store_test_data_file_to_db(data)

        # Store data to test cases for reporting
        if 'media' not in self.dut.test_cases[self.name]:
            self.dut.test_cases[self.name]['media'] = []

        self.dut.test_cases[self.name]['media'].append(data)

    def _store_test_data_file_to_db(self, data):
        if self.db_handler:
            self.db_handler.store_test_data_file_to_db(**data)



"""Store any additional test parameters to PARAMETERS dictionary.
When pool of test cases is used, PARAMETERS can be used to store
DUT type specific parameters"""

PARAMETERS = {"param1": "this is param1", "param2": "and this is param2"}



"""Initializations and deinitializations"""


def boot_up(instruments, logger):
    logger.info("Executing initialization")


def finalize_test(overallresult, duts, instruments, logger):
    logger.info("Testing ready. Release DUT(s) on Gaia")

    if overallresult == 'pass':
        instruments['gaia'].state_triggers['ReleasePass']()
    else:
        instruments['gaia'].state_triggers['ReleaseFail']()

    instruments['gaia'].wait_not_ready()


def prepare_test(instruments, logger, dut_sn_values, sequence_name):
    logger.info("Waiting G5 to get ready for next (or first) test run")
    instruments['gaia'].wait_ready()


def identify_DUTs(dut_sn, instruments, logger):
    import uuid

    # Figure out DUT sn (probably with code reader), find out what
    # sequence must be used with the DUT type.
    # Return DUTs and sequence name.
    # Sequence name is the directory name at test_definitions
    # Here we use testA as an example. Thus you must create
    # the sequence with 'python super_simple_test_runner.py -c testA'
    return (
        # DUTs
        {
            "1": {'sn': str(uuid.uuid4())},
            "2": {'sn': str(uuid.uuid4())},
            "3": {'sn': str(uuid.uuid4())},
        },
        # Test sequence
        "testA",
        # Operator info
        {"name": "John Doe"},
    )


def shutdown(instruments, logger):
    logger.info("Shutdown")



"""Instrument definitions"""
import time
import gaiaclient
import requests
import pymongo


# This definition will also run when new test sequence is created.
# Thus it is not good idea to connect to instrument here
# connect at instrument_initialization instead.

INSTRUMENTS = {'gaia': None, 'test_data_db': None}


def get_tester_info():
    # Tester info
    return {"name": "First delivery", "type": "G5", "serial": "1234abc"}


def handle_instrument_status(progress, logger):
    """Reconnect to instrument, do other error resolving or any other status triggered task.
    Called in the beginning of each test run."""

    def _connect_to_db(logger):
        # This import doesn't work when creating new test sequence
        # so let's import it here when needed
        from common import mongodb_handler
        try:
            INSTRUMENTS['test_data_db'] = mongodb_handler.DatabaseHandler(
                'mongodb://JOTUser:YOURPASSWORDHERE@localhost:2701/production_testing',
                'production_testing',
            )
            INSTRUMENTS['Line test_data_db'].db_client.server_info()

        except pymongo.errors.OperationFailure as err:
            return err.details['errmsg']
        except pymongo.errors.ServerSelectionTimeoutError:
            return 'Connection error'
        else:
            return "OK"

    _connect_to('test_data_db', progress, _connect_to_db, logger)

    def _connect_to_gaia(logger):
        try:
            INSTRUMENTS['gaia'] = gaiaclient.Client(gaiaclient.Client("http://localhost:1234"))

        except requests.exceptions.ConnectionError as e:
            return "Connection error"
        else:
            return "OK"

    _connect_to('gaia', progress, _connect_to_gaia, logger)


def _connect_to(name, progress, connector, logger):
    def get_status():
        try:
            progress.set_instrument_status(name, "Trying to connect...")

            return connector(logger)
        except Exception as e:
            logger.exception(f'Failed to connect to {name}', e)
            return 'error'

    if progress.instrument_status[name] != "MagicMock":
        if progress.instrument_status[name] != "OK":

            status = get_status()

            while status != 'OK':

                time.sleep(1)
                # Set status after sleep so that the user sees "Trying to connect"
                # This is handy is connector() doesn't give meaningful response
                # then the user still sees the "trying to connect" message
                progress.set_instrument_status(name, status)
                status = get_status()

            progress.set_instrument_status(name, status)



import json


class ProgressReporter:
    def __init__(self, test_control, progess_queue):
        self.test_control = test_control
        self.statistics = None
        self.general_step = None
        self.test_positions = None
        self.overall_result = None
        self.sequence_name = None
        self.operator_instructions = None
        self.progess_queue = progess_queue
        self.instrument_status = None
        self.version_info = None

    def set_instrument_status(self, instrument, status):

        if self.instrument_status is None:
            self.instrument_status = {}

        self.instrument_status[instrument] = status
        self._report_progress()

    def set_version_info(self, version_name, version_value):

        if self.version_info is None:
            self.version_info = {}

        self.version_info[version_name] = version_value
        self._report_progress()

    def set_progress(self, **kwargs):
        # Store variable
        self.__dict__.update(kwargs)
        # Send new progress json
        self._report_progress()

    def show_operator_instructions(self, message, append=False):
        if append and self.operator_instructions:
            self.operator_instructions = self.operator_instructions + '\r\n' + message
        else:
            self.operator_instructions = message

        self._report_progress()

    def _report_progress(self):

        positions_dict = {}

        if self.test_positions:
            for position_name, position_value in self.test_positions.items():
                dut_class = None
                if position_value.dut is None:
                    if position_value.previous_dut:
                        dut_class = position_value.previous_dut
                else:
                    dut_class = position_value.dut.get_dut_dict()
                positions_dict[position_name] = {
                    'step': position_value.step,
                    'status': position_value.status,
                    'sn': None if position_value.dut is None else position_value.dut.serial_number,
                    'test_status': str(position_value.test_status),
                    'dut_class': dut_class
                }

        progress_json = {
            "general_state": self.general_step,
            "duts": positions_dict,
            "sequence_name": self.sequence_name,
            "get_sn_from_ui": self.test_control['get_sn_from_ui'],
            "test_sequences": self.test_control['test_sequences'],
            "test_cases": self.test_control['test_cases'],
        }

        if self.overall_result:
            progress_json['overall_result'] = self.overall_result

        progress_json['statistics'] = self.statistics
        progress_json['instrument_status'] = self.instrument_status
        progress_json['version_info'] = self.version_info

        progress_json['operator_instructions'] = self.operator_instructions

        self.test_control['progress'] = progress_json
        self.progess_queue.put(json.dumps(progress_json, default=str))



class TestPosition:
    def __init__(self, name, ui_position):
        self.name = name
        self.ui_position = ui_position
        self.step = None
        self.status = None
        self.test_status = None
        self.test_case_instances = {}
        self.dut = None
        self.stop_testing = False
        self.previous_dut = False

    def prepare_for_new_test_run(self):
        if self.dut:
            self.previous_dut = self.dut.get_dut_dict()

        self.step = None
        self.status = 'wait'
        self.dut = None
        self.test_case_instances = {}
        self.stop_testing = False

        if not self.test_status:
            self.test_status = "Idle"

    def __str__(self):
        return self.name



from setuptools import setup, find_packages
import glob
import os
import tarfile
import pathlib


"""
To publish:
1. Make sure you have all the dependencies (Activate virtual env. Import Docker will fail silently, if missing!)
2. Set correct version number to this file
3. python setup.py bdist_wheel
4. twine upload dist/*

"""


def get_ui():
    try:
        import docker
    except ImportError:
        return
    file_stream = open('ui.tar', 'wb')
    client = docker.from_env()
    client.images.pull('ci.jot.local:5000/iris_ui')
    ctnr = client.containers.create('ci.jot.local:5000/iris_ui', name='iris_ui')
    api_client = docker.APIClient()
    bits, stat = api_client.get_archive('iris_ui', '/usr/src/app/build')

    for chunk in bits:
        file_stream.write(chunk)
    file_stream.close()
    ctnr.remove()
    tar = tarfile.open("ui.tar")
    tar.extractall('ui')
    tar.close()

    # Remove tar file
    pathlib.Path('ui.tar').unlink()


def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


with open("README.md", "r") as fh:
    long_description = fh.read()

get_ui()

setup(
    name="jot-iris",
    version="1.18.4",
    license="MIT License",
    author="JOT Automation Ltd.",
    author_email="rami.rahikkala@jotautomation.com",
    description="Super simple test sequencer for production testing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.jotautomation.com",
    packages=find_packages(),
    # Package_data will add files to package(s) defined in previous file.
    # Here we add files to ui/build and to additional_dist_files
    # This way we get UI and other non-python files distributed
    package_data={
        "ui": package_files("ui/build/"),
        "additional_dist_files": package_files('additional_dist_files/'),
    },
    scripts=["iris.py"],
    install_requires=[
        "wheel",
        "json2html",
        "tornado",
        "gaiaclient",
        "PyYAML",
        "colorlog",
        "colorama",
        "mock",
        "pymongo",
        "bcrypt"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""
Handles the creation of the result files and directories
"""

import datetime
import pathlib
import json2html


def create_report(
    report_json,
    report_dict,
    test_positions,
    parameters,
    local_db,
    common_definitions,
    progress_reporter,
):
    """Creates and stores report for DUT(s)"""

    progress_reporter.show_operator_instructions("Put DUT to yellow box!")

    # Get root level items on report
    root_items = {}

    for key, value in report_dict.items():
        if not isinstance(value, dict) or key == 'operator' or key == 'tester':
            root_items[key] = value

    # Extract each DUT and add to database
    for key, value in test_positions.items():

        dut = value.dut

        if dut is not None:
            dut_sn = dut.serial_number

            result_db = {'serialnumber': dut_sn}

            result_db['result'] = dut.pass_fail_result

            result_db['failedCases'] = dut.failed_steps

            result_db['passedCases'] = [
                case_name for case_name, case in dut.test_cases.items() if case['result']
            ]
            # The actual results
            result_db['testCases'] = report_dict[dut_sn]

            # Add also root level items
            result_db.update(root_items)
            local_db.db_client[local_db.db_name].test_reports.insert_one(result_db)

    filename = '_'.join(
        [
            '%s-%s' % (test_position_name, test_position_value.dut.serial_number)
            for (test_position_name, test_position_value) in test_positions.items()
        ]
    )

    report_path = create_report_path()

    report_file_path = report_path / (filename + '.html')

    report_file_path.write_text(json2html.json2html.convert(json=report_json))


def create_report_path():
    current = datetime.datetime.now()

    report_path = (
        pathlib.Path.cwd() / 'results' / str(current.year) / str(current.month) / str(current.day)
    )
    report_path.mkdir(parents=True, exist_ok=True)

    return report_path



"""Runs the test cases"""
import datetime
import time
import json
import threading
import logging
from unittest.mock import MagicMock
from test_runner import progress_reporter
from test_runner import helpers
from test_runner import exceptions


def get_common_definitions():
    """Returns test definitions"""

    return helpers.import_by_name(
        'common',
        "No test definitions defined? Create definition template with --create argument.",
        logging.getLogger("common_definitions"),
    )


def get_test_cases(logger):

    test_cases = {}
    for sequence in get_common_definitions().TEST_SEQUENCES:
        test_definitions = helpers.get_test_definitions(sequence, logger)
        test_cases[sequence] = [
            {'name': t}
            for t in test_definitions.TESTS
            if t not in test_definitions.SKIP and '_pre' not in t and '_post' not in t
        ]
    return test_cases


def get_test_control(logger):
    """Returns default test control dictionary"""
    return {
        'single_run': False,
        'step': False,
        'skip': None,
        'loop_all': False,
        'loop': None,
        'retest_on_fail': 0,
        'terminate': False,
        'abort': False,
        'report_off': False,
        'run': threading.Event(),
        'get_sn_from_ui': get_common_definitions().SN_FROM_UI,
        'test_sequences': get_common_definitions().TEST_SEQUENCES,
        'dry_run': False,
        'test_cases': get_test_cases(logger),
    }


def get_sn_from_ui(dut_sn_queue, logger):
    """Returns serial numbers from UI"""

    sequence_name = None
    test_cases = None
    common_definitions = get_common_definitions()
    duts_sn = {
        test_position.name: {'sn': None} for test_position in common_definitions.TEST_POSITIONS
    }
    logger.info(
        'Wait SNs from UI for test_positions: '
        + ", ".join([str(t) for t in common_definitions.TEST_POSITIONS])
    )

    while True:
        msg = dut_sn_queue.get()

        try:
            msg = json.loads(msg)
            for dut in msg:
                if dut in duts_sn:
                    duts_sn[dut]['sn'] = msg[dut]
            if 'sequence' in msg:
                sequence_name = msg['sequence']
            if 'testCases' in msg and msg['testCases']:

                test_cases = [t['name'] for t in msg['testCases']]

        except (AttributeError, json.decoder.JSONDecodeError):
            pass

        # Loop until all test_positions have received a serial number
        for dut in duts_sn:
            if not duts_sn[dut]['sn']:
                break
        else:
            logger.info("All DUT serial numbers received from UI")
            logger.info("Selected test %s", sequence_name)
            break

    return (duts_sn, sequence_name, {"name": "Not available"}, test_cases)


def run_test_runner(test_control, message_queue, progess_queue, dut_sn_queue, listener_args):
    """Starts the testing"""

    logger = logging.getLogger('test_runner')

    def send_message(message):
        if message:
            message_queue.put(message)

    common_definitions = get_common_definitions()

    progress = progress_reporter.ProgressReporter(test_control, progess_queue)

    progress.set_progress(general_state="Boot")

    for instrument in common_definitions.INSTRUMENTS.keys():
        progress.set_instrument_status(instrument, 'Not initialized')

    if test_control['dry_run']:

        for instrument in common_definitions.INSTRUMENTS.keys():
            common_definitions.INSTRUMENTS[instrument] = MagicMock()
            progress.set_instrument_status(instrument, 'MagicMock')

    elif 'mock' in test_control:

        for instrument in common_definitions.INSTRUMENTS.keys():
            if instrument in test_control['mock']:
                common_definitions.INSTRUMENTS[instrument] = MagicMock()
                progress.set_instrument_status(instrument, 'MagicMock')

    elif 'inverse_mock' in test_control:

        for instrument in common_definitions.INSTRUMENTS.keys():
            if instrument not in test_control['inverse_mock']:
                common_definitions.INSTRUMENTS[instrument] = MagicMock()
                progress.set_instrument_status(instrument, 'MagicMock')

    logger.info("Initializing instruments")

    # Initialize all instruments
    common_definitions.handle_instrument_status(progress, logger)

    logger.info("All instruments initialized")

    db_handler = common_definitions.INSTRUMENTS[common_definitions.DB_HANDLER_NAME]

    listener_args['database'] = db_handler

    # Execute boot_up defined for the test sequence
    common_definitions.boot_up(common_definitions.INSTRUMENTS, logger)

    test_positions = {}
    fail_reason_history = ''
    fail_reason_count = 0
    pass_count = 0

    for position in common_definitions.TEST_POSITIONS:
        test_positions[position.name] = position

    # Start the actual test loop
    while not test_control['terminate']:
        # Wait until you are allowed to run again i.e. pause
        test_control['run'].wait()
        logger.info("Start new test run")

        try:
            background_pre_tasks = {}
            background_post_tasks = []
            test_control['abort'] = False

            logger.info("Checking status of instruments")

            progress.set_progress(
                general_state="Checking status of instruments",
                overall_result=None,
                test_positions=test_positions,
            )

            logger.info("All instruments OK")

            common_definitions.handle_instrument_status(progress, logger)

            if db_handler:
                db_handler.clean_db()
                if isinstance(db_handler, MagicMock):
                    progress.set_progress(statistics={'statistics': 'mocked'})
                else:
                    progress.set_progress(statistics=db_handler.get_statistics())

            # Clear test positions
            for position_name, position in test_positions.items():
                position.prepare_for_new_test_run()

                progress.set_progress(general_state="Prepare", test_positions=test_positions)

            test_cases_override = None

            # DUT sn may come from UI
            if test_control['get_sn_from_ui']:

                (
                    dut_sn_values,
                    sequence_name,
                    operator_info,
                    test_cases_override,
                ) = get_sn_from_ui(dut_sn_queue, logger)

                sequence_name_from_identify = common_definitions.identify_DUTs(
                    dut_sn_values, common_definitions.INSTRUMENTS, logger
                )

                # If sequence was not selected, get it from identify_DUTs
                if sequence_name is None:
                    sequence_name = sequence_name_from_identify

            else:
                # Or from identify_DUTs function
                (dut_sn_values, sequence_name, operator_info,) = common_definitions.identify_DUTs(
                    None, common_definitions.INSTRUMENTS, logger
                )

            common_definitions.prepare_test(
                common_definitions.INSTRUMENTS, logger, dut_sn_values, sequence_name
            )
            # Create dut instances
            for test_position, dut_info in dut_sn_values.items():
                if dut_info is None:
                    test_positions[test_position].dut = None
                else:
                    test_positions[test_position].dut = common_definitions.parse_dut_info(
                        dut_info, test_position
                    )

            results = {"operator": operator_info, "tester": common_definitions.get_tester_info()}

            # Fetch test definitions i.e. import module
            test_definitions = helpers.get_test_definitions(sequence_name, logger)

            # Fetch test case pool too
            test_pool = helpers.get_test_pool_definitions(logger)

            # Remove skipped test_case_names from test list
            test_case_names = [t for t in test_definitions.TESTS if t not in test_definitions.SKIP]

            start_time_epoch = time.time()
            start_time = datetime.datetime.now()
            start_time_monotonic = time.monotonic()
            test_run_id = str(start_time_epoch).replace('.', '_')

            # Run all test cases
            for test_case_name in test_case_names:
                # Loop for testing

                if test_control['abort']:

                    send_message("Test aborted")
                    logger.warning("Test aborted")
                    break

                if test_cases_override and test_case_name not in test_cases_override:
                    continue

                is_pre_test = False
                if '_pre' in test_case_name:
                    is_pre_test = True

                test_case_name = test_case_name.replace('_pre', '').replace('_pre', '')

                # Run test cases for each DUT in test position
                for test_position_name, test_position_instance in test_positions.items():

                    # Set sn to be none, if you don't want to run any test_case_names for the test position
                    # but want to keep showing the position on the UI
                    if not test_position_instance.dut or test_position_instance.stop_testing:
                        continue

                    # Fill DUT data
                    test_position_instance.step = test_case_name
                    test_position_instance.status = 'testing'

                    progress.set_progress(
                        general_state='testing',
                        test_positions=test_positions,
                        sequence_name=sequence_name,
                    )

                    def new_test_instance(the_case, the_position_instance):
                        if hasattr(test_definitions, the_case):
                            test_instance = getattr(test_definitions, the_case)(
                                test_definitions.LIMITS,
                                progress,
                                the_position_instance.dut,
                                test_definitions.PARAMETERS,
                                db_handler,
                                common_definitions,
                            )
                        elif hasattr(test_pool, the_case):
                            test_instance = getattr(test_pool, the_case)(
                                test_definitions.LIMITS,
                                progress,
                                the_position_instance.dut,
                                test_definitions.PARAMETERS,
                                db_handler,
                                common_definitions,
                            )
                        else:
                            raise exceptions.TestCaseNotFound(
                                "Cannot find specified test case: " + the_case
                            )
                        return test_instance

                    # Create test case instance
                    if test_case_name not in test_position_instance.test_case_instances:

                        test_position_instance.test_case_instances[
                            test_case_name
                        ] = new_test_instance(test_case_name, test_position_instance)
                    test_instance = test_position_instance.test_case_instances[test_case_name]
                    test_instance.test_position = test_position_instance
                    test_instance.test_run_id = test_run_id
                    try:
                        if is_pre_test:
                            # Start pre task and store it to dictionary
                            if test_case_name not in background_pre_tasks:
                                background_pre_tasks[test_case_name] = {}

                            background_pre_tasks[test_case_name][
                                test_position_name
                            ] = threading.Thread(target=test_instance.run_pre_test)
                            background_pre_tasks[test_case_name][test_position_name].start()
                        else:
                            # Wait for pre task
                            if (
                                test_case_name in background_pre_tasks
                                and test_position_name in background_pre_tasks[test_case_name]
                            ):
                                background_pre_tasks[test_case_name][test_position_name].join()
                            else:
                                # Or if pre task is not run, run it now
                                test_instance.run_pre_test()

                            test_position_instance.test_status = "Testing"
                            progress.set_progress(
                                general_state='testing',
                                test_position=test_positions,
                                sequence_name=sequence_name,
                            )
                            # Run the actual test case
                            test_instance.run_test()

                            progress.set_progress(
                                general_state='testing',
                                test_positions=test_positions,
                                sequence_name=sequence_name,
                            )
                            test_position_instance.test_status = "Idle"

                            # Start post task and store it to list

                            bg_task = threading.Thread(target=test_instance.run_post_test)
                            bg_task.start()

                            background_post_tasks.append(bg_task)

                    except Exception as err:

                        trace = []
                        trace_back = err.__traceback__
                        while trace_back is not None:
                            trace.append(
                                {
                                    "filename": trace_back.tb_frame.f_code.co_filename,
                                    "name": trace_back.tb_frame.f_code.co_name,
                                    "line": trace_back.tb_lineno,
                                }
                            )
                            trace_back = trace_back.tb_next

                        err_dict = {
                            'type': type(err).__name__,
                            'message': str(err),
                            'trace': trace,
                        }

                        test_instance.handle_error(error=err_dict)

                    else:
                        # No error and no active tests
                        test_position_instance.status = 'idle'
                        test_position_instance.step = None

                        progress.set_progress(
                            general_state='testing',
                            test_positions=test_positions,
                            sequence_name=sequence_name,
                        )

            for task in background_post_tasks:
                task.join()

            for test_position_name, test_position_instance in test_positions.items():

                dut = test_position_instance.dut
                results[dut.serial_number] = test_position_instance.dut.test_cases

                if not test_position_instance.dut:
                    continue

                if dut.pass_fail_result == 'error':
                    errors = [
                        f"{case_name}: {case['error']}"
                        for case_name, case in dut.test_cases.items()
                        if case['result'] == 'error' and 'error' in case
                    ]

                    send_message(f"{dut.serial_number}: ERROR: " + ', '.join(errors))
                    test_position_instance.test_status = 'error'
                elif dut.pass_fail_result == 'pass':
                    send_message(f"{dut.serial_number}: PASSED")
                    test_position_instance.test_status = 'pass'
                    pass_count = pass_count + 1
                else:
                    send_message(f"{dut.serial_number}: FAILED: {', '.join(dut.failed_steps)}")
                    test_position_instance.test_status = 'fail'

                    if fail_reason_history == dut.failed_steps:
                        fail_reason_count = fail_reason_count + 1
                    else:
                        fail_reason_count = 0
                        fail_reason_history = dut.failed_steps
                    pass_count = 0

            if fail_reason_count > 4 and pass_count < 5:
                send_message(f"WARNING: 5 or more consecutive fails on {fail_reason_history}")

            progress.set_progress(
                general_state='finalize',
                test_positions=test_positions,
                overall_result=dut.pass_fail_result,
                sequence_name=sequence_name,
            )
            common_definitions.finalize_test(
                dut.pass_fail_result, test_positions, common_definitions.INSTRUMENTS, logger
            )

            results["start_time"] = start_time
            results["start_time_epoch"] = start_time_epoch
            results["end_time"] = datetime.datetime.now()
            results["test_run_id"] = test_run_id

            results["duration_s"] = round(time.monotonic() - start_time_monotonic, 2)

        except exceptions.IrisError as e:
            # TODO: write error to report
            logger.exception("Error on testsequence")
            continue
        except Exception as exp:
            # TODO: write error to report
            logger.exception("Error on testsequence")
            continue

        else:
            pass
        finally:
            pass

        # Don't create report if aborted
        if test_control['abort']:
            common_definitions.test_aborted(common_definitions.INSTRUMENTS, logger)
            continue

        progress.set_progress(
            general_state="Create test report",
            test_positions=test_positions,
            overall_result=dut.pass_fail_result,
            sequence_name=sequence_name,
        )
        try:
            if not test_control['report_off']:
                common_definitions.create_report(
                    json.dumps(results, indent=4, default=str),
                    results,
                    test_positions,
                    test_definitions.PARAMETERS,
                    db_handler,
                    common_definitions,
                    progress,
                )
        except Exception as e:
            progress.set_progress(general_state="Error")
            send_message("Error while generating a test report")
            send_message(str(e))

        if test_control['single_run']:
            test_control['terminate'] = True

    progress.set_progress(general_state="Shutdown")

    common_definitions.shutdown(common_definitions.INSTRUMENTS, logger)
